#!/usr/bin/env python3

"""Corn jobs 🌽

Supporting the core app with asynchronous maintenance tasks.

Mostly related to monitoring and managing Printer state.
"""

from time import sleep
from threading import Thread, Event
from typing import Callable
from datetime import timedelta, datetime
import logging
from contextlib import closing
from urllib import parse as urlparse
from tentacles.store import Store

from octorest import OctoRest as _OR
from requests import Response
from requests.exceptions import HTTPError, ConnectionError, Timeout


class OctoRest(_OR):
    def _get(self, path, params=None):
        url = urlparse.urljoin(self.url, path)
        response = self.session.get(url, params=params, timeout=(0.1, 1.0))
        self._check_response(response)

        return response.json()

    def _check_response(self, response: Response):
        response.raise_for_status()
        return response


log = logging.getLogger(__name__)

SHUTDOWN = Event()


def corn_job(every: timedelta):
    def _decorator(f):
        def _helper(*args, **kwargs):
            last = None
            while not SHUTDOWN.is_set():
                if not last or (datetime.now() - last) > every:
                    log.debug(f"Ticking job {f.__name__}")
                    try:
                        last = datetime.now()
                        f(*args, **kwargs)
                    except Exception:
                        log.exception(f"Error while procesing task {f.__name__}")
                else:
                    sleep(1)

        return _helper

    return _decorator


@corn_job(timedelta(seconds=5))
def poll_printers(db_factory: Callable[[], Store]) -> None:
    """Poll printers for their status."""

    with closing(db_factory()) as db:
        for printer in db.list_printers():
            id, name, url, api_key, last_poll, status = printer
            try:
                client = OctoRest(url=url, apikey=api_key)
                job = client.job_info()
                state = client.printer().get("state").get("flags", {})
                if state.get("error"):
                    db.update_printer_status(id, "error")
                elif state.get("ready"):
                    db.update_printer_status(id, "idle")
                elif state.get("printing"):
                    db.update_printer_status(id, "running")
                else:
                    raise Exception(f"Indeterminate state {state!r}")

            except (ConnectionError, Timeout):
                db.update_printer_status(id, "error")

            except HTTPError as e:
                assert isinstance(e.response, Response)
                if e.response.status_code in [403, 401] or "error" in job:
                    db.update_printer_status(id, "error")
                elif e.response.json().get("error") == "Printer is not operational":
                    db.update_printer_status(id, "disconnected")
                else:
                    raise Exception(
                        f"Indeterminate state {e.response.status_code}, {e.response.json()!r}"
                    )


@corn_job(timedelta(seconds=5))
def assign_jobs(db_factory: Callable[[], Store]) -> None:
    """Assign jobs to printers. Uploading files and job state management is handled separately."""

    with closing(db_factory()) as db:
        for printer_id in db.list_idle_printers():
            printer = db.fetch_printer(printer_id)
            if printer.status != "idle":
                continue

            if next_job_id := db.poll_job_queue():
                db.assign_job(next_job_id, printer_id)


@corn_job(timedelta(seconds=5))
def push_jobs(db_factory: Callable[[], Store]) -> None:
    """Ensure that job files are uploaded and started to the assigned printer."""

    with closing(db_factory()) as db:
        for job in db.list_mapped_jobs():
            printer = db.fetch_printer(job.printer_id)
            file = db.fetch_file(job.file_id)
            try:
                client = OctoRest(printer.url, printer.api_key)
                client.upload(file.filename, select=True, print=True)
                db.start_job(job.id)
            except Exception:
                log.exception("Oop")


@corn_job(timedelta(seconds=5))
def pull_jobs(db_factory: Callable[[], Store]) -> None:
    """Poll the state of mapped printers to control jobs."""

    with closing(db_factory()) as db:
        for job in db.list_running_jobs():
            printer = db.fetch_printer(job.printer_id)
            if printer.status != "running":
                db.finish_job(job.id, printer.status)


def create_workers(db_factory: Callable[[], Store]) -> Event:
    Thread(target=poll_printers, args=[db_factory]).start()
    Thread(target=assign_jobs, args=[db_factory]).start()
    Thread(target=push_jobs, args=[db_factory]).start()
    Thread(target=pull_jobs, args=[db_factory]).start()

    return SHUTDOWN
